import numpy as np
from collections import OrderedDict

from pykin.kinematics.transform import Transform
from pykin.kinematics import transformation as tf
from pykin.kinematics import jacobian as jac


class Kinematics:
    def __init__(self, tree):
        self.tree = tree

    @staticmethod
    def _forward_kinematics(root, theta_dict, offset=Transform()):
        link_transforms = OrderedDict()
        trans = offset * root.get_transform(theta_dict.get(root.joint.name, 0.0))
        link_transforms[root.link.name] = trans * root.link.offset
        for child in root.children:
            link_transforms.update(
                Kinematics._forward_kinematics(child, theta_dict, trans)
            )

        return link_transforms

    def forward_kinematics(self, thetas, offset=Transform(), desired_tree=None):
        if desired_tree is None:
            if not isinstance(thetas, dict):
                joint_names = self.tree.get_joint_parameter_names
                assert len(joint_names) == len(
                    thetas
                ), f"the number of joints is {len(joint_names)}, but the number of joint's angle is {len(thetas)}"
                thetas_dict = dict((j, thetas[i]) for i, j in enumerate(joint_names))
            else:
                thetas_dict = thetas
            return self._forward_kinematics(self.tree.root, thetas_dict, offset)
        else:
            cnt = 0
            link_transforms = {}
            trans = offset
            for f in desired_tree:
                trans = trans * f.get_transform(thetas[cnt])
                link_transforms[f.link.name] = trans * f.link.offset
                if f.joint.dtype != "fixed":
                    cnt += 1
                if cnt >= len(thetas):
                    cnt -= 1
            return link_transforms

    def analytical_inverse_kinematics(self, pose):
        # Link Length [m]
        L = [0.27035, 0.069, 0.36435, 0.069, 0.37429, 0.0, 0.38735]  # 0.36830
        L_h = 0.37082  # np.sqrt(L[2] ** 2 + L[3] ** 2)


        T_BL_0 = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, L[0]], [0, 0, 0, 1]])
        T_BR_0 = T_BL_0

        TM_6_GL = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, L[6]], [0, 0, 0, 1]])
        TM_6_GR = TM_6_GL

        # pose to Transformation matrix
        T_BL_Goal = tf.pose_to_homogeneous(pose)

        T_0_6 = np.dot(np.dot(np.linalg.inv(T_BL_0), T_BL_Goal), np.linalg.inv(TM_6_GL))

        # position
        x = T_0_6[0, 3]
        y = T_0_6[1, 3]
        z = T_0_6[2, 3]

        # calculate theta 1
        theta_1 = np.arctan2(y, x)

        # Parameters for calculating theta 2
        E = 2 * L_h * (L[1] - (x / np.cos(theta_1)))
        F = 2 * L_h * z
        G = (
            ((x ** 2) / (np.cos(theta_1) ** 2))
            + (L[1] ** 2 + L_h ** 2 - L[4] ** 2)
            + (z ** 2)
            - ((2 * L[1] * x) / (np.cos(theta_1)))
        )

        k = E ** 2 + F ** 2 - G ** 2
        print(k)
        if k > 0:
            k = np.sqrt(k)
        else:
            print("k is lower 0")
            k = 0

        # tangent 1, 2 for calculating theta 2
        tan_1 = (-F + k) / (G - E)
        tan_2 = (-F - k) / (G - E)

        # calcuate theta_2_1, theta_2_2
        theta_2_1 = 2 * np.arctan(tan_1)
        theta_2_2 = 2 * np.arctan(tan_2)

        theta_3 = 0

        # calcuate theta 4_1, 4_2
        theta_4_1 = (
            np.arctan2(
                (-z - L_h * np.sin(theta_2_1)),
                ((x / np.cos(theta_1) - L[1] - L_h * np.cos(theta_2_1))),
            )
            - theta_2_1
        )
        theta_4_2 = (
            np.arctan2(
                (-z - L_h * np.sin(theta_2_2)),
                ((x / np.cos(theta_1) - L[1] - L_h * np.cos(theta_2_2))),
            )
            - theta_2_2
        )
        R_0_6 = T_0_6[0:3, 0:3]

        theta_2_4_1 = theta_2_1 + theta_4_1
        R_0_3_1 = np.array(
            [
                [
                    -np.cos(theta_1) * np.sin(theta_2_4_1),
                    -np.cos(theta_1) * np.cos(theta_2_4_1),
                    -np.sin(theta_1),
                ],
                [
                    -np.sin(theta_1) * np.sin(theta_2_4_1),
                    -np.sin(theta_1) * np.cos(theta_2_4_1),
                    np.cos(theta_1),
                ],
                [-np.cos(theta_2_4_1), np.sin(theta_2_4_1), 0],
            ]
        )
        R_3_6_1 = np.dot(np.matrix.transpose(R_0_3_1), R_0_6)

        theta_2_4_2 = theta_2_2 + theta_4_2
        R_0_3_2 = np.array(
            [
                [
                    -np.cos(theta_1) * np.sin(theta_2_4_2),
                    -np.cos(theta_1) * np.cos(theta_2_4_2),
                    -np.sin(theta_1),
                ],
                [
                    -np.sin(theta_1) * np.sin(theta_2_4_2),
                    -np.sin(theta_1) * np.cos(theta_2_4_2),
                    np.cos(theta_1),
                ],
                [-np.cos(theta_2_4_2), np.sin(theta_2_4_2), 0],
            ]
        )
        R_3_6_2 = np.dot(np.matrix.transpose(R_0_3_2), R_0_6)

        theta_5_1 = np.arctan2(R_3_6_1[2, 2], R_3_6_1[0, 2])
        theta_5_2 = np.arctan2(R_3_6_2[2, 2], R_3_6_2[0, 2])

        theta_7_1 = np.arctan2(-R_3_6_1[1, 1], R_3_6_1[1, 0])
        theta_7_2 = np.arctan2(-R_3_6_2[1, 1], R_3_6_2[1, 0])

        theta_6_1 = np.arctan2((R_3_6_1[1, 0] / np.cos(theta_7_1)), -R_3_6_1[1, 2])
        theta_6_2 = np.arctan2((R_3_6_2[1, 0] / np.cos(theta_7_2)), -R_3_6_2[1, 2])

        result1 = [
            theta_1,
            theta_2_1,
            theta_3,
            theta_4_1,
            theta_5_1,
            theta_6_1,
            theta_7_1,
        ]
        result2 = [
            theta_1,
            theta_2_2,
            theta_3,
            theta_4_2,
            theta_5_2,
            theta_6_2,
            theta_7_2,
        ]

        return result1, result2

    # TODO
    # singularity problem
    def numerical_inverse_kinematics(self, current_joints, target, desired_tree=None):

        lamb = 0.5
        iterator = 1
        maxIter = 10000
        EPS = float(1e-6)
        dof = len(current_joints)

        def rot_to_omega(R):
            
            el = np.array(
                [[R[2, 1] - R[1, 2]], 
                 [R[0, 2] - R[2, 0]], 
                 [R[1, 0] - R[0, 1]]]
            )
            norm_el = np.linalg.norm(el)
            if norm_el > EPS:
                w = np.dot(np.arctan2(norm_el, np.trace(R) - 1) / norm_el, el)
            elif (R[0, 0] > 0 and R[1,1] > 0 and R[2, 2] > 0):
                w = np.zeros((3, 1))
            else:
                w = np.dot(np.pi/2, np.array([[R[0,0] + 1], [R[1,1] + 1], [R[2,2] + 1]]))
            return w

        def calc_pose_error(T_ref, T_cur):

            pos_err = np.array([T_ref[:3, -1] - T_cur[:3, -1]])
            rot_err = np.dot(T_cur[:3, :3].T, T_ref[:3, :3])
            w_err = np.dot(T_cur[:3, :3], rot_to_omega(rot_err))

            return np.vstack((pos_err.T, w_err))

        # Step 1. Prepare the position and attitude of the target link
        target_pose = tf.get_homogeneous_matrix(target[:3], target[3:])

        # Step 2. Use forward kinematics to calculate the position and attitude of the target link
        cur_fk = self.forward_kinematics(current_joints, desired_tree=desired_tree)
        cur_pose = list(cur_fk.values())[-1].matrix()

        # Step 3. Calculate the difference in position and attitude
        err_pose = calc_pose_error(target_pose, cur_pose)
        err = np.linalg.norm(err_pose)

        # Step 4. If error is small enough, stop the calculation
        while err > EPS:
            # Step 5. If error is not small enough, calculate dq which would reduce the error 
            # Get jacobian to calculate dq 
            J = jac.calc_jacobian(desired_tree, cur_fk, current_joints)
            dq = lamb * np.dot(np.linalg.pinv(J), err_pose)

            # Step 6. Update joint angles by q = q + dq and calculate forward Kinematics
            current_joints = [current_joints[i] + dq[i] for i in range(dof)]

            cur_fk = self.forward_kinematics(current_joints, desired_tree=desired_tree)
            cur_pose = list(cur_fk.values())[-1].matrix()
            err_pose = calc_pose_error(target_pose, cur_pose)
            err = np.linalg.norm(err_pose)

            # Avoid infinite calculation
            iterator += 1
            if iterator > maxIter:
                break

        current_joints = [float(current_joint) for current_joint in current_joints]
        return current_joints



import numpy as np
from xml.etree import ElementTree as ET
from collections import OrderedDict
from pprint import pprint

from pykin.urdf.urdf_tree import URDFTree
from pykin.geometry.frame import Joint, Link, Frame
from pykin.kinematics.transform import Transform

JOINT_TYPE_MAP = {'revolute': 'revolute',
                  'fixed': 'fixed',
                  'prismatic': 'prismatic'}


def _convert_transform(origin):
    if origin is None:
        return Transform()
    else:
        return Transform(rot=origin.rot, pos=origin.pos)

def _convert_string_to_narray(str_input):
    return np.array([float(data) for data in str_input.split()])

class URDFParser:
    def __init__(self, filepath=None):
        if filepath:
            self.parse(filepath)

    def parse(self, filepath):
        tree_xml = ET.parse(filepath)
        root = tree_xml.getroot()
        robot_name = root.attrib.get('name')

        tree = URDFTree(name=robot_name)

        for idx, link_tag in enumerate(root.findall('link')):
            link_frame = self._parse_link(link_tag, idx=idx)
            tree.links[link_frame.link.name] = link_frame.link

        for idx, joint_tag in enumerate(root.findall('joint')):
            joint_frame = self._parse_joint(joint_tag, idx=idx)
            tree.joints[joint_frame.joint.name] = joint_frame.joint

        root_name = next(iter(tree.links))
        tree.root = tree.links[root_name]

        root_frame = Frame(root_name + "_frame")
        root_frame.joint = Joint()
        root_frame.link = Link(root_name, offset=_convert_transform(root_frame.link.offset))
        root_frame.children = self._build_chain_recursive(tree.root, tree.links, tree.joints)

        tree.root = root_frame
        self.tree = tree
        return tree

    def _parse_link(self, link_tag, idx):
        attrib = link_tag.attrib
        link_name = attrib.get('name', 'link_' + str(idx))
        frame = Frame(link_name + '_frame',
                      link=Link(link_name, offset=Transform()))

        return frame

    def _parse_joint(self, joint_tag, idx):
        attrib = joint_tag.attrib
        joint_name = attrib.get('name', 'joint_' + str(idx))

        frame = Frame(joint_name + '_frame',
                      joint=Joint(name=joint_name, offset=Transform(), dtype=attrib['type']))

        parent_tag = joint_tag.find('parent')
        frame.joint.parent = parent_tag.attrib['link']

        child_tag = joint_tag.find('child')
        frame.joint.child = child_tag.attrib['link']

        # origin
        origin_tag = joint_tag.find('origin')
        if origin_tag is not None:
            frame.joint.offset.pos = _convert_string_to_narray(origin_tag.attrib.get('xyz'))
            frame.joint.offset.rot = _convert_string_to_narray(origin_tag.attrib.get('rpy'))

        # axis
        axis_tag = joint_tag.find('axis')
        if axis_tag is not None:
            frame.joint.axis = _convert_string_to_narray(axis_tag.attrib.get('xyz'))

        return frame

    def _build_chain_recursive(self, root: Link, links: OrderedDict, joints: OrderedDict) -> list:
        children = []
        for joint in joints.values():

            if joint.parent == root.name:
                child_frame = Frame(joint.child + "_frame")
                child_frame.joint = Joint(joint.name, offset=_convert_transform(joint.offset), dtype=JOINT_TYPE_MAP[joint.dtype], axis=joint.axis)

                chil_link = links[joint.child]
                child_frame.link = Link(chil_link.name, offset=_convert_transform(chil_link.offset))
                child_frame.children = self._build_chain_recursive(child_frame.link, links, joints)
                children.append(child_frame)

        return children




import sys, os
import numpy as np
from pprint import pprint
pykin_path = os.path.abspath(os.path.dirname(__file__)+"../" )
sys.path.append(pykin_path)

from pykin.kinematics import jacobian as jac
from pykin.kinematics.kinematics import Kinematics
from pykin.kinematics.transform import Transform
from pykin.urdf.urdf_parser import URDFParser
from pykin.utils import plot as plt
from pykin.utils.shell_color import ShellColors as scolors


class Robot:
    def __init__(self, filepath=None, offset=Transform()):
        if filepath is None:
            filepath = pykin_path + "/asset/urdf/baxter.urdf"
        self._offset = offset
        self.tree = None
        self.desired_frame = None
        self._load_urdf(filepath)

    def __repr__(self):
        return f"""ROBOT : {__class__.__name__} 
        {self.links} 
        {self.joints}"""

    def _load_urdf(self, filepath):
        parser = URDFParser(filepath)
        self.tree = parser.tree
        self.tree.offset = self.offset
        self.__kinematics = Kinematics(self.tree)

    @property
    def links(self):
        links = []
        for link in self.tree.links.values():
            links.append(link)
        return links

    @property
    def joints(self):
        joints = []
        for joint in self.tree.joints.values():
            joints.append(joint)
        return joints

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, offset):
        self._offset = offset

    @property
    def num_dofs(self):
        return self.tree.num_dofs

    @property
    def num_links(self):
        return self.tree.num_links

    @property
    def num_joints(self):
        return self.tree.num_joints

    @property
    def num_active_joints(self):
        return self.tree.num_actuated_joints

    @property
    def get_active_joint_names(self):
        return self.tree.get_joint_parameter_names

    def show_robot_info(self):
        print("*" * 20)
        print(f"robot information: \n{self}")
        print(f"robot's dof : {self.num_dofs}")
        print(f"active joint's info: \n{self.get_active_joint_names}")
        print("*" * 20)

    def set_desired_tree(self, root_link="", end_link=""):
        self.desired_frame = self.tree._set_desired_tree(root_link, end_link)
        return self.desired_frame

    def forward_kinematics(self, theta, desired_tree=None):
        self.transformations = self.__kinematics.forward_kinematics(
            theta, offset=self._offset, desired_tree=self.desired_frame
        )
        return self.transformations

    def inverse_kinematics(
        self, current_joints, target_pose, method="numerical", desired_tree=None
    ):
        if method == "analytical":
            return self.__kinematics.analytical_inverse_kinematics(target_pose)
        if method == "numerical":
            return self.__kinematics.numerical_inverse_kinematics(
                current_joints, target_pose, desired_tree=self.desired_frame
            )

    def jacobian(self, fk, th):
        return jac.calc_jacobian(self.desired_frame, fk, th)


if __name__ == "__main__":
    file_path = str(pykin_path) + "/asset/urdf/baxter.urdf"
    if len(sys.argv) > 1:
        file_path = sys.argv[1]

    robot = Robot(file_path)
    robot.show_robot_info()
    # print(robot)
    ##################################################################
    ## Forward Kinematics
    head_thetas = [0.0]
    right_arm_thetas = [0, 0, 0, 0, 0, 0, 0]
    left_arm_thetas = [0, 0, 0, 0, 0, 0, 0]

    # # left_arm_thetas = [0, -np.pi/6, np.pi, -np.pi, 0, -np.pi/6, 0]

    # for jacobian
    robot.set_desired_tree("base", "left_gripper")
    print(robot.tree.get_desired_joint_parameter_names())

    fk = robot.forward_kinematics(left_arm_thetas)
    position, orientation = fk["left_gripper"].pos, fk["left_gripper"].rot
    pose = np.append(position, orientation)
    print(fk)
    # J = robot.jacobian(fk, left_arm_thetas)
    # print(J)
    # pose = [0.06402554 , 1.18271238, 0.320976,   0.49999954, - 0.50000046 , 0.49999954 ,0.50000046]
    # pose = [0.06402611,  0.87651951, - 0.26023899 , 0.2705978, - 0.65328208  ,0.65328088 ,0.2705983]
    pose = [
        0.06402868,
        -0.52665762,
        0.478976,
        -0.49999954,
        -0.50000046,
        0.49999954,
        -0.50000046,
    ]
    result = robot.inverse_kinematics(left_arm_thetas, pose, method="numerical")

    thethas = head_thetas + right_arm_thetas + result
    print(result)
    robot.desired_frame = None
    _, ax = plt.init_3d_figure()
    plt.plot_robot(robot, thethas, ax, "baxter")
    ax.legend()
    plt.show_figure()
    ####################
    ########################################################################

    # thetas = head_thetas + right_arm_thetas + left_arm_thetas
    # transformations = robot.forward_kinematics(thetas)

    # position, orientation = transformations["left_gripper"].pos, transformations["left_gripper"].rot
    # pose = np.append(position, orientation)

    # ## Inverse Kinematics
    # print(f"target pose : {pose}")

    # _, ax = plt.init_3d_figure()
    # plt.plot_robot(robot, thetas, ax, "baxter")
    # ax.legend()
    # plt.show_figure()

    # result1, result2 = robot.inverse_kinematics(pose, method="analytical")
    # thethas = head_thetas + right_arm_thetas + result1
    # print(result1)
    # _, ax = plt.init_3d_figure()
    # plt.plot_robot(robot, thethas, ax, "baxter")
    # ax.legend()
    # plt.show_figure()

    # print(result2)
    # thethas = head_thetas + right_arm_thetas + result2
    # _, ax = plt.init_3d_figure()
    # plt.plot_robot(robot, thethas, ax, "baxter")
    # ax.legend()
    # plt.show_figure()
    ######################################################

    ######################################################
    head_thetas = [0.0]
    right_arm_thetas = [0, 0, 0, 0, 0, 0, 0]
    left_arm_thetas = [np.pi / 4, np.pi, 0, 0, 0, 0, 0]
    thetas = head_thetas + right_arm_thetas + left_arm_thetas
    transformations = robot.forward_kinematics(thetas)
    position, orientation = (
        transformations["left_wrist"].pos,
        transformations["left_wrist"].rot,
    )
    pose = np.append(position, orientation)
    print(pose)
    _, ax = plt.init_3d_figure()
    plt.plot_robot(robot, thetas, ax, "baxter")
    ax.legend()
    plt.show_figure()
    ####################################################

    # # # To calculate right arm transformations
    # head_thetas = [0.0]
    # right_arm_thetas = [0, 0, 0, 0, 0, 0, 0]
    # left_arm_thetas = [0, 0, 0, 0, 0, 0, 0]

    # thetas = head_thetas + right_arm_thetas + left_arm_thetas
    # robot.forward_kinematics(thetas)
    # active_joints_transformation = robot.get_active_joint_transform()
    # right_links = list(active_joints_transformation)[1:8]

    # right_joint_transforms = []
    # for link, (joint, transform) in active_joints_transformation.items():
    #     for right_link in right_links:
    #         if link == right_link:
    #             right_joint_transforms.append(transform)

    # # print(right_joint_transforms)
    # homogeneous_matrixes = []
    # for right_joint_transform in right_joint_transforms:
    #     T = tf.get_homogeneous_matrix(right_joint_transform.pos, right_joint_transform.rot)
    #     homogeneous_matrixes.append(np.round(T, decimals=5))
    # print(homogeneous_matrixes)
    # _, ax = plt.init_3d_figure()
    # plt.plot_right_arm(robot, homogeneous_matrixes, ax, "test")
    # plt.show_figure()
    # ####################################################

    # ####################################################
    ## iiwa14
    # thetas = [0, 0, 0, 0, 0, 0, 0]

    # transformations = robot.forward_kinematics(thetas)

    # _, ax = plt.init_3d_figure()
    # plt.plot_robot(robot, thetas, ax, "test")
    # ax.legend()
    # plt.show_figure()
    # ####################################################

    # print(robot.get_active_joint_names)
    # print(robot.num_dofs)
    # th = np.zeros(robot.num_dofs)



import numpy as np


def calc_jacobian(desired_frame, fk: dict, thetas: list) -> np.array:
    jsize = len(thetas)
    target_position = list(fk.values())[-1].pos
    J = np.zeros((6, jsize))
    n = 0
    for frame in desired_frame:
        if frame.joint.dtype == "revolute":
            n += 1
            w = np.dot(fk[frame.link.name].matrix()[:3, :3], frame.joint.axis)
            v = np.cross(w, target_position - fk[frame.link.name].pos)
            J[:, n - 1] = np.hstack((v, w))
        elif frame.joint.dtype == "prismatic":
            n += 1
            w = np.zeros(3)
            v = np.dot(fk[frame.link.name].matrix()[:3, :3], frame.joint.axis)
            J[:, n - 1] = np.hstack((v, w))
    return J



import os
import numpy as np
import matplotlib.animation
import matplotlib.pyplot as plt
from pykin.kinematics import transformation as tf
# Colors of each directions axes. For ex X is green
directions_colors = ["green", "cyan", "orange"]


def plot_basis(robot, ax, arm_length=1):
    """Plot a frame fitted to the robot size"""

    offset = np.linalg.norm(robot.offset.pos)

    if offset == 0:
        offset = 1

    ax.set_xlim3d([-1.0 * offset, 1.0 * offset])
    ax.set_xlabel('X')

    ax.set_ylim3d([-1.0 * offset, 1.0 * offset])
    ax.set_ylabel('Y')

    ax.set_zlim3d([-1.0 * offset, 1.0 * offset])
    ax.set_zlabel('Z')

    # Plot du repère
    # Sa taille est relative à la taille du bras
    ax.plot([0, offset * 1.5], [0, 0], [0, 0],
            c=directions_colors[0], label="X")
    ax.plot([0, 0], [0, offset * 1.5], [0, 0],
            c=directions_colors[1], label="Y")
    ax.plot([0, 0], [0, 0], [0, offset * 1.5],
            c=directions_colors[2], label="Z")


def plot_robot(robot, fk, ax, name=None):
    # TODO
    # nodes : list to dict (name, value)
    if name is not None:
        name = os.path.splitext(os.path.basename(name))[0].strip()

    plot_basis(robot, ax)
    links = []
    nodes = []
    transformation_matrix = []

    for (link, transformation) in fk.items():
        links.append(link)
        transformation_matrix.append(
            tf.get_homogeneous_matrix(transformation.pos, transformation.rot))

    for link, matrix in zip(links, transformation_matrix):
        nodes.append(tf.get_pos_mat_from_homogeneous(matrix))

    if name == "baxter":
        torso_nodes = [nodes[0], nodes[3]]
        head_nodes = torso_nodes + nodes[7:12]
        pedestal_nodes = torso_nodes + [nodes[6]]
        right_nodes = torso_nodes + nodes[13:18] + nodes[20:29]
        left_nodes = torso_nodes + nodes[31:36] + nodes[38:47]

        head_lines = ax.plot([x[0] for x in head_nodes], [x[1] for x in head_nodes], [
            x[2] for x in head_nodes], linewidth=5, label="head")
        pedestal_lines = ax.plot([x[0] for x in pedestal_nodes], [x[1] for x in pedestal_nodes], [
            x[2] for x in pedestal_nodes], linewidth=5, label="pedestal")
        right_lines = ax.plot([x[0] for x in right_nodes], [x[1] for x in right_nodes], [
            x[2] for x in right_nodes], linewidth=5, label="right arm")
        left_lines = ax.plot([x[0] for x in left_nodes], [x[1] for x in left_nodes], [
            x[2] for x in left_nodes], linewidth=5, label="left arm")

        head_label = '(%0.4f, %0.4f, %0.4f)' % (
            head_nodes[-1][0], head_nodes[-1][1], head_nodes[-1][2])
        pedestal_label = '(%0.4f, %0.4f, %0.4f)' % (
            pedestal_nodes[-1][0], pedestal_nodes[-1][1], pedestal_nodes[-1][2])
        right_label = '(%0.4f, %0.4f, %0.4f)' % (
            right_nodes[8][0], right_nodes[8][1], right_nodes[8][2])
        left_label = '(%0.4f, %0.4f, %0.4f)' % (
            left_nodes[8][0], left_nodes[8][1], left_nodes[8][2])

        ax.text(head_nodes[-1][0], head_nodes[-1][1],
                head_nodes[-1][2], head_label, size="8")
        ax.text(pedestal_nodes[-1][0], pedestal_nodes[-1][1],
            pedestal_nodes[-1][2], pedestal_label, size="8")
        ax.text(right_nodes[-1][0], right_nodes[-1][1],
                right_nodes[-1][2], right_label, size="8")
        ax.text(left_nodes[-1][0], left_nodes[-1][1],
                left_nodes[-1][2], left_label, size="8")

        ax.scatter([x[0] for x in head_nodes], [x[1] for x in head_nodes], 
            [x[2] for x in head_nodes], s=55, c=head_lines[0].get_color())
        ax.scatter([x[0] for x in pedestal_nodes], [x[1] for x in pedestal_nodes], 
            [x[2] for x in pedestal_nodes], s=55, c=pedestal_lines[0].get_color())
        ax.scatter([x[0] for x in right_nodes], [x[1] for x in right_nodes], 
            [x[2] for x in right_nodes], s=55, c=right_lines[0].get_color())
        ax.scatter([x[0] for x in left_nodes], [x[1] for x in left_nodes], 
            [x[2] for x in left_nodes], s=55, c=left_lines[0].get_color())
    else:
        lines = ax.plot([x[0] for x in nodes], [x[1] for x in nodes], [
            x[2] for x in nodes], linewidth=5, label=name)

        label = '(%0.4f, %0.4f, %0.4f)' % (
            nodes[-1][0], nodes[-1][1], nodes[-1][2])

        ax.text(nodes[-1][0], nodes[-1][1],
                nodes[-1][2], label, size="8")
        
        ax.scatter([x[0] for x in nodes], [x[1] for x in nodes],
            [x[2] for x in nodes], s=55, c=lines[0].get_color())

    
def init_3d_figure():
    from mpl_toolkits.mplot3d import axes3d, Axes3D
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    return fig, ax


def show_figure():
    plt.show()





import numpy as np
from collections import OrderedDict

from pykin.kinematics import transform as tf
from pykin.kinematics import transformation


class URDFTree:
    def __init__(self, name=None, offset=tf.Transform(), root=None):
        self.name = name
        self.offset = offset
        self.links = OrderedDict()
        self.joints = OrderedDict()
        self.root = root
        self.desired_root = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, offset):
        self._offset = offset

    @property
    def num_dofs(self):
        return sum([joint.num_dof for joint in self.joints.values()])

    @property
    def num_links(self):
        return len(self.links)

    @property
    def num_joints(self):
        return len(self.joints)

    @property
    def num_fixed_joints(self):
        return sum([1 for joint in self.joints.values() if joint.num_dof == 0])

    @property
    def num_actuated_joints(self):
        return sum([1 for joint in self.joints.values() if joint.num_dof != 0])

    def find_frame(self, name):
        if self.root.name == name:
            return self.root
        return self._find_frame_recursive(name, self.root)

    @staticmethod
    def _find_frame_recursive(name, frame):
        for child in frame.children:
            if child.name == name:
                return child
            ret = URDFTree._find_frame_recursive(name, child)
            if not ret is None:
                return ret
        return None

    def find_link(self, name):
        if self.root.link.name == name:
            return self.root.link
        return self._find_link_recursive(name, self.root)

    @staticmethod
    def _find_link_recursive(name, frame):
        for child in frame.children:
            if child.link.name == name:
                return child.link
            ret = URDFTree._find_link_recursive(name, child)
            if not ret is None:
                return ret
        return None

    @property
    def get_joint_parameter_names(self):
        joint_names = []
        joint_names = self._get_joint_parameter_names(joint_names, self.root)
        return joint_names

    def _get_joint_parameter_names(self, joint_names, frame):
        if frame.joint.num_dof != 0:
            joint_names.append(frame.joint.name)
        for child in frame.children:
            self._get_joint_parameter_names(joint_names, child)
        return joint_names

    def _set_desired_tree(self, root_link_name="", end_link_name=""):
        if root_link_name == "":
            self.desired_root = self.root
        else:
            self.desired_root = self.find_frame(root_link_name + "_frame")
        self.desired_frame = self._generate_desired_tree_recursive(
            self.desired_root, end_link_name
        )
        if self.desired_root is None:
            raise ValueError("Invalid end frame name %s." % end_link_name)
        self.desired_frame = [self.desired_root] + self.desired_frame
        return self.desired_frame

    @staticmethod
    def _generate_desired_tree_recursive(root_frame, end_link_name):
        for child in root_frame.children:
            if child.link.name == end_link_name:
                return [child]
            else:
                frames = URDFTree._generate_desired_tree_recursive(child, end_link_name)
                if frames is not None:
                    return [child] + frames

    def get_desired_joint_parameter_names(self):
        names = []
        for f in self.desired_frame:
            if f.joint.dtype == "fixed":
                continue
            names.append(f.joint.name)
        return names



class RobotParser(object):
    
    def __init__(self, filepath=None):
        self.root = None
        self.tree = None
        self.filepath = filepath
        if filepath is not None:
            self.parse(filepath)

    @property
    def root(self):
        return self._root
    
    @root.setter
    def root(self, root):
        self._root = root
    
    @property
    def tree(self):
        return self._tree

    @tree.setter
    def tree(self, tree):
        self._tree = tree

    def parse(self, filepath):
        pass

    



import numpy as np
import math
from collections import Iterable

def vector_norm(data, axis=None, out=None):
    """Return length, i.e. Euclidean norm, of ndarray along axis.

    >>> v = numpy.random.random(3)
    >>> n = vector_norm(v)
    >>> numpy.allclose(n, numpy.linalg.norm(v))
    True
    >>> v = numpy.random.rand(6, 5, 3)
    >>> n = vector_norm(v, axis=-1)
    >>> numpy.allclose(n, numpy.sqrt(numpy.sum(v*v, axis=2)))
    True
    >>> n = vector_norm(v, axis=1)
    >>> numpy.allclose(n, numpy.sqrt(numpy.sum(v*v, axis=1)))
    True
    >>> v = numpy.random.rand(5, 4, 3)
    >>> n = numpy.empty((5, 3))
    >>> vector_norm(v, axis=1, out=n)
    >>> numpy.allclose(n, numpy.sqrt(numpy.sum(v*v, axis=1)))
    True
    >>> vector_norm([])
    0.0
    >>> vector_norm([1])
    1.0

    """
    data = np.array(data, dtype=np.float64, copy=True)
    if out is None:
        if data.ndim == 1:
            return math.sqrt(np.dot(data, data))
        data *= data
        out = np.atleast_1d(np.sum(data, axis=axis))
        np.sqrt(out, out)
        return out
    data *= data
    np.sum(data, axis=axis, out=out)
    np.sqrt(out, out)
    return None


def get_rot_mat_from_homogeneous(homogeneous_matrix):
    return homogeneous_matrix[:-1, :-1]

# TODO
# do change name
def get_pos_mat_from_homogeneous(homogeneous_matrix):
    return homogeneous_matrix[:-1,-1]


def get_pose_from_homogeneous(homogeneous_matrix):
    position = get_pos_mat_from_homogeneous(homogeneous_matrix)
    orientation = get_quaternion_from_matrix(
        get_rot_mat_from_homogeneous(homogeneous_matrix))
    return np.hstack((position, orientation))


def get_rpy_from_matrix(R):
    r = np.arctan2(R[2, 1], R[2, 2])
    p = np.arctan2(-R[2, 0], np.sqrt(R[0, 0] ** 2 + R[1, 0] ** 2))
    y = np.arctan2(R[1, 0], R[0, 0])

    return np.asarray([r, p, y])


def get_rpy_from_quaternion(q, convention='wxyz'):
    if convention == 'xyzw':
        x, y, z, w = q[0], q[1], q[2], q[3]  # (N,)
    elif convention == 'wxyz':
        w, x, y, z = q[0], q[1], q[2], q[3]  # (N,)
    roll = np.arctan2(2 * (w*x + y*z), 1 - 2 * (x**2 + y**2))   # (N,)
    pitch = np.arcsin(2 * (w*y - z*x))                          # (N,)
    yaw = np.arctan2(2 * (w*z + x*y), 1 - 2 * (y**2 + z**2))    # (N,)
    rpy = np.asarray([roll, pitch, yaw]).T  # (N,3)
    return rpy

def get_matrix_from_rpy(rpy):
    cr, cp, cy = [np.cos(i) for i in rpy]
    sr, sp, sy = [np.sin(i) for i in rpy]
    R = np.array([[cy*cp, cy*sp*sr - sy*cr, cy*sp*cr + sy*sr],
                  [sy*cp, sy*sp*sr + cy*cr, sy*sp*cr - cy*sr],
                  [-sp, cp*sr, cp*cr]])
    return R


def get_matrix_from_axis_angle(axis, angle):
    x, y, z = axis
    theta = angle
    c, s = np.cos(theta), np.sin(theta)
    v = 1 - c
    R = np.array([[x ** 2 * v + c, x * y * v - z * s, x * z * v + y * s],
                  [x * y * v + z * s, y ** 2 * v + c, y * z * v - x * s],
                  [x * z * v - y * s, y * z * v + x * s, z ** 2 * v + c]])
    return R


def get_matrix_from_quaternion(q, convention='wxyz'):
    if isinstance(q, Iterable):
        if convention == 'xyzw':
            x, y, z, w = q
        elif convention == 'wxyz':
            w, x, y, z = q
    else:
        raise TypeError
    R = np.array([[2 * (w**2 + x**2) - 1, 2 * (x*y - w*z), 2 * (x*z + w*y)],
                  [2 * (x*y + w*z), 2 * (w**2 + y**2) - 1, 2*(y*z - w*x)],
                  [2 * (x*z - w*y), 2 * (y*z + w*x), 2 * (w**2 + z**2) - 1]])
    return R


def get_quaternion_from_rpy(rpy, convention='wxyz'):
    rpy = np.asarray(rpy)
    multiple_rpy = True
    if len(rpy.shape) < 2:
        multiple_rpy = False
        rpy = np.array([rpy])  # (1,3)

    r, p, y = rpy[:, 0], rpy[:, 1], rpy[:, 2]
    cr, sr = np.cos(r/2.), np.sin(r/2.)
    cp, sp = np.cos(p/2.), np.sin(p/2.)
    cy, sy = np.cos(y/2.), np.sin(y/2.)

    w = cr * cp * cy + sr * sp * sy  # (N,)
    x = sr * cp * cy - cr * sp * sy  # (N,)
    y = cr * sp * cy + sr * cp * sy  # (N,)
    z = cr * cp * sy - sr * sp * cy  # (N,)

    if convention == 'xyzw':
        q = np.vstack([x, y, z, w]).T
    elif convention == 'wxyz':
        q = np.vstack([w, x, y, z]).T
    else:
        raise NotImplementedError(
            "Asking for a convention that has not been implemented")

    if not multiple_rpy:
        return q[0]
    return q


def get_quaternion_from_matrix(R, convention='wxyz'):
    w = 1./2 * np.sqrt(R[0, 0] + R[1, 1] + R[2, 2] + 1)
    x, y, z = 1./2 * np.array([np.sign(R[2, 1] - R[1, 2]) * np.sqrt(R[0, 0] - R[1, 1] - R[2, 2] + 1),
                               np.sign(R[0, 2] - R[2, 0]) *
                               np.sqrt(R[1, 1] - R[2, 2] - R[0, 0] + 1),
                               np.sign(R[1, 0] - R[0, 1]) * np.sqrt(R[2, 2] - R[0, 0] - R[1, 1] + 1)])

    if convention == 'xyzw':
        return np.array([x, y, z, w])
    elif convention == 'wxyz':
        return np.array([w, x, y, z])
    else:
        raise NotImplementedError(
            "Asking for a convention that has not been implemented")


def get_quaternion_from_axis_angle(axis, angle, convention='wxyz'):
    w = np.cos(angle / 2.)
    x, y , z = np.sin(angle / 2.) * axis
    if convention == 'xyzw':
        return np.array([x, y, z, w])
    elif convention == 'wxyz':
        return np.array([w, x, y, z])
    else:
        raise NotImplementedError("Asking for a convention that has not been implemented")


def get_quaternion_inverse(quaternion):
    """Return inverse of quaternion.

    >>> q0 = random_quaternion()
    >>> q1 = quaternion_inverse(q0)
    >>> numpy.allclose(quaternion_multiply(q0, q1), [1, 0, 0, 0])
    True

    """
    q = np.array(quaternion, dtype=np.float64, copy=True)
    np.negative(q[1:], q[1:])
    return q / np.dot(q, q)


def quaternion_matrix(quaternion):
    """Return homogeneous rotation matrix from quaternion.

    >>> M = quaternion_matrix([0.99810947, 0.06146124, 0, 0])
    >>> numpy.allclose(M, rotation_matrix(0.123, [1, 0, 0]))
    True
    >>> M = quaternion_matrix([1, 0, 0, 0])
    >>> numpy.allclose(M, numpy.identity(4))
    True
    >>> M = quaternion_matrix([0, 1, 0, 0])
    >>> numpy.allclose(M, numpy.diag([1, -1, -1, 1]))
    True

    """
    q = np.array(quaternion, dtype=np.float64, copy=True)
    n = np.dot(q, q)
    if n < _EPS:
        return np.identity(4)
    q *= math.sqrt(2.0 / n)
    q = np.outer(q, q)
    return np.array([
        [1.0 - q[2, 2] - q[3, 3], q[1, 2] - q[3, 0], q[1, 3] + q[2, 0], 0.0],
        [q[1, 2] + q[3, 0], 1.0 - q[1, 1] - q[3, 3], q[2, 3] - q[1, 0], 0.0],
        [q[1, 3] - q[2, 0], q[2, 3] + q[1, 0], 1.0 - q[1, 1] - q[2, 2], 0.0],
        [0.0, 0.0, 0.0, 1.0]])


def quaternion_multiply(quaternion1, quaternion0):
    """Return multiplication of two quaternions.

    >>> q = quaternion_multiply([4, 1, -2, 3], [8, -5, 6, 7])
    >>> numpy.allclose(q, [28, -44, -14, 48])
    True

    """
    w0, x0, y0, z0 = quaternion0
    w1, x1, y1, z1 = quaternion1
    return np.array(
        [
            -x1 * x0 - y1 * y0 - z1 * z0 + w1 * w0,
            x1 * w0 + y1 * z0 - z1 * y0 + w1 * x0,
            -x1 * z0 + y1 * w0 + z1 * x0 + w1 * y0,
            x1 * y0 - y1 * x0 + z1 * w0 + w1 * z0,
        ],
        dtype=np.float64,
    )

# epsilon for testing whether a number is close to zero
_EPS = np.finfo(float).eps * 4.0

def get_quaternion_about_axis(angle, axis):
    """Return quaternion for rotation about axis.

    >>> q = quaternion_about_axis(0.123, [1, 0, 0])
    >>> numpy.allclose(q, [0.99810947, 0.06146124, 0, 0])
    True

    """
    q = np.array([0.0, axis[0], axis[1], axis[2]])
    qlen = vector_norm(q)
    if qlen > _EPS:
        q *= math.sin(angle / 2.0) / qlen
    q[0] = math.cos(angle / 2.0)
    return q


def get_homogeneous_matrix(position, orientation):
    position = np.asarray(position)
    orientation = np.asarray(orientation)
    if orientation.shape == (3,):  # RPY Euler angles
        R = get_matrix_from_rpy(orientation)
    elif orientation.shape == (4,):  # quaternion in the form [x,y,z,w]
        R = get_matrix_from_quaternion(orientation)
    elif orientation.shape == (3, 3):  # Rotation matrix
        R = orientation

    H = np.vstack((np.hstack((R, position.reshape(-1, 1))),
                  np.array([[0, 0, 0, 1]])))
    return H


def get_inverse_homogeneous(matrix):
    R = matrix[:3, :3].T
    p = -R.dot(matrix[:3, 3].reshape(-1,1))
    return np.vstack((np.hstack((R,p)),
                      np.array([[0, 0, 0, 1]])))


def get_identity_homogeneous_matrix():
    return np.identity(4)


def homogeneous_to_pose(matrix):
    position = matrix[:3, -1]
    quaternion = get_quaternion_from_matrix(matrix[:3, :3])
    return np.concatenate((position, quaternion))


def pose_to_homogeneous(pose):
    pose = np.array(pose).flatten()
    position, orientation = pose[:3], pose[3:]
    return get_homogeneous_matrix(position=position, orientation=orientation)


def get_quaternion(orientation, convention='wxyz'):
    if isinstance(orientation, tuple) and len(orientation) == 2:
        angle, axis = orientation
        if isinstance(axis, (float, int)) and isinstance(angle, np.ndarray):
            angle, axis = axis, angle
        return get_quaternion_from_axis_angle(axis, angle, convention)
    if isinstance(orientation, (np.ndarray, list)):
        orientation = np.asarray(orientation)
        if orientation.shape == (3,):
            return get_quaternion_from_rpy(orientation, convention)
        if orientation.shape == (4,):
            if convention == 'wxyz':
                x, y, z, w = orientation
                return np.array([w, x, y, z])
            return orientation
        if orientation.shape == (3, 3):
            return get_quaternion_from_matrix(orientation, convention)
        else:
            raise ValueError("Expecting the shape of the orientation to be (3,), (3,3), or (4,), instead got: "
                             "{}".format(orientation.shape))
    else:
        raise TypeError("Expecting the given orientation to be a np.ndarray, quaternion, tuple or list, instead got: "
                        "{}".format(type(orientation)))


def get_rotation_matrix(orientation):
    if isinstance(orientation, tuple) and len(orientation) == 2:
        angle, axis = orientation
        if isinstance(axis, (float, int)) and isinstance(angle, np.ndarray):
            angle, axis = axis, angle
        return get_matrix_from_axis_angle(axis, angle)
    if isinstance(orientation, (np.ndarray, list)):
        orientation = np.asarray(orientation)
        if orientation.shape == (3,):  # RPY Euler angles
            return get_matrix_from_rpy(orientation)
        if orientation.shape == (4,):  # quaternion
            return get_matrix_from_quaternion(orientation)
        if orientation.shape == (3, 3):  # rotation matrix
            return orientation
        else:
            raise ValueError("Expecting the shape of the orientation to be (3,), (3,3), or (4,), instead got: "
                         "{}".format(orientation.shape))
    else:
        raise TypeError("Expecting the given orientation to be a np.ndarray, quaternion, tuple or list, instead got: "
                    "{}".format(type(orientation)))
